/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz> (original author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <math.h>
#include <nspr.h>
/* #include <fenv.h> */
#include <string.h>

#define IN_INTEGRATOR

#define CYCLE_FUNCTION do_integrator_cycle

#include "SpecialMaths.c"
#include "IntegratorCommand.h"
#include "IntegrationHelperHead.c"

static void
cmd_start(void *data, PRUint32 length)
{
  struct StartIntegrationCommand *cmd;
  PRUint32 i;
  if (length < sizeof(struct StartIntegrationCommand))
    return;
  cmd = (struct StartIntegrationCommand*)data;

  /* Zero out lots of things... */
#define Z(x) memset(x, 0, sizeof(x))
  Z(zz);
  Z(scal);
  Z(ff);
  Z(jac);
  Z(e1);
  Z(e2r);
  Z(e2i);
  Z(cont);
  Z(ip1);
  Z(ip2);
  Z(iphes);
#undef Z

  /* Copy initial values across... */
  memcpy(
	 all_yvalues,
	 cmd->initialValues,
	 sizeof(cmd->initialValues)
	);
  memcpy(
         constants,
         cmd->constantValues,
         sizeof(cmd->constantValues)
        );

  EvaluateSimple(x, all_yvalues);

  rel_tolerance = cmd->rel_tolerance;
  abs_tolerance = cmd->abs_tolerance;
  resolution = cmd->resolution;
  xend = cmd->stopX;
  xpoint = x = cmd->startX;

  /* Do some initialisations */
  unexp = PR_FALSE;
  unexn = PR_FALSE;
  change = PR_FALSE;
  current_order = MINIMAL_ORDER;
  n_current_order = MAXIMAL_ORDER * ND;
  ikeep = 0;
  ichan = 0;
  theta = 0;
  thetat = 0;
  lrc = n_current_order + ND;
  do_coercv(current_order);
  posneg = ((xend - x) >= 0) ? 1 : -1;
  maximal_step_size = abs_d(xend - x);
  step_size = INITIAL_STEP_SIZE;  
  if (step_size > maximal_step_size)
    step_size = maximal_step_size;
  step_size *= posneg;
  step_size_old = step_size;
  reject = PR_FALSE;
  first = PR_TRUE;
  last = PR_FALSE;
  caljac = PR_FALSE;
  skipjac = PR_FALSE;
  calhes = PR_FALSE;
  newton_count = 0;

  /* If they try a very small range, we do it in one step... */
  if ((x + step_size*1.0001 - xend) * posneg > 0.0)
  {
    step_size = xend - x;
    last = PR_TRUE;
  }
  step_size_opt = step_size;
  faccon = 1.0;
  xold = x;
  expo = 1.0 / (current_order + 1.0);
  expmns = (current_order + 1.0)/(2 * current_order);
  quott = abs_tolerance / rel_tolerance;
  rel_tolerance_1 = 0.1 * pow(rel_tolerance, expmns);
  abs_tolerance_1 = rel_tolerance_1 * quott;
  hhfac = step_size;
  nstep = 0;
  naccpt = 0;

  for (i = 0; i < ND; i++)
    scal[i] = abs_tolerance_1 + rel_tolerance_1 * abs_d(all_yvalues[i + NS]);

  /* Add the first point... */
  EvaluateSimple(x, all_yvalues);
#if 0
  add_points(x, all_yvalues);
#endif

  /* Evaluate the differentials into y_0... */
  ComputeDifferentials(x, all_yvalues, y_0);

  helper_state = RUNNING;
}

static void
cmd_suspend(void *data, PRUint32 length)
{
  struct IntegratorReplyHeader irh;
  helper_state = SUSPENDED;
  irh.length = sizeof(struct IntegratorReplyHeader);
  irh.reply = INTEGRATION_HALTED;
#ifdef DEBUG_MODE
  printf("Sending integration halted reply.\n");
#endif
  send_packet((char*)&irh, sizeof(irh), PR_FALSE);
}

static void
cmd_stop(void *data, PRUint32 length)
{
  helper_state = DYING;
}

static inline void
integration_done()
{
  struct IntegratorReplyHeader irh;
  helper_state = SUSPENDED;
  irh.length = sizeof(struct IntegratorReplyHeader);
  irh.reply = INTEGRATION_DONE;
#ifdef DEBUG_MODE
  printf("Sending integration_done packet.\n");
#endif
  send_packet((char*)&irh, sizeof(irh), PR_FALSE);
#ifdef DEBUG_MODE
  printf("Queue cleared\n");
#endif
}

typedef struct
{
  struct IntegratorReplyHeader rh;
  double x;
  double y[NS + ND];
} PointPacket;

static inline void
add_points(double xval, double* values)
{
  while (xpoint <= xval)
  {
    PointPacket* pkt = (PointPacket*)PR_Malloc(sizeof(PointPacket));
    pkt->rh.length = sizeof(*pkt);
    pkt->rh.reply = INTEGRATION_POINT;
    pkt->x = xpoint;
    do_interpolation(xpoint, pkt->y);
    xpoint += resolution;
#ifdef DEBUG_MODE
    printf("Queueing Point Packet for x=%f\n", pkt->x);
    printf("First variable is %f\n", pkt->y[0]);
#endif
    queue_packet((char*)pkt, sizeof(*pkt), PR_TRUE);
  }
}

static inline void
integration_failed(const char* error)
{
  char *data;
  PRUint32 l;
  struct IntegratorReplyHeader* irh;
  l = sizeof(struct IntegratorReplyHeader) + strlen(error) + 1;
  data = (char*)PR_Malloc(l);
  helper_state = SUSPENDED;
  irh = (struct IntegratorReplyHeader*)data;
  irh->length = l;
  irh->reply = INTEGRATION_FAILED;
  strcpy(data + sizeof(struct IntegratorReplyHeader), error);
#ifdef DEBUG_MODE
  printf("Sending integration failed packet: %s\n", error);
#endif
  send_packet(data, l, PR_TRUE);
}
